@@ -2,5 +2,6 @@ from toolformer_pytorch.palm import PaLM
 
 from toolformer_pytorch.toolformer_pytorch import (
     Toolformer,
-    filter_tokens_with_api_response
+    filter_tokens_with_api_response,
+    sample
 )
